# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements a JATIC compatible ART PyTorchClassifier.
"""
from dataclasses import dataclass
from typing import Any, Sequence

from art.estimators.classification.pytorch import PyTorchClassifier
from maite.errors import InvalidArgument
from maite.protocols import HasLogits, SupportsArray

from heart_library.utils import process_inputs_for_art

META_NOT_SPECIFIED: str = "Not specified"


@dataclass
class JaticPytorchClassifierOutput:
    """
    Dataclass output of ART JATIC PyTorch Classifier
    """

    logits: SupportsArray


@dataclass
class HeartClassifierMetadata:
    """
    HEART metadata dataclass for PyTorch Classifier
    """

    model_name: str
    provider: str
    task: str = "Image Classification"


class JaticPyTorchClassifier(PyTorchClassifier):  # pylint: disable=R0901
    """
    JATIC compatible extension of ART core PyTorchClassifier
    """

    metadata: HeartClassifierMetadata

    def __init__(
        self,
        labels: Sequence[str],
        model_name: str = META_NOT_SPECIFIED,
        provider: str = META_NOT_SPECIFIED,
        **kwargs: Any
    ):
        super().__init__(**kwargs)
        if labels is None:
            raise InvalidArgument("No labels were provided.")
        self._labels: Sequence[str] = labels
        self.metadata = HeartClassifierMetadata(model_name, provider)

    def __call__(self, data: SupportsArray) -> HasLogits:

        # convert to ART supported type
        images, _ = process_inputs_for_art(data, self._device)

        # make prediction
        output = self.predict(images)

        # convert back to JATIC supported type
        return JaticPytorchClassifierOutput(output)

    def get_labels(self) -> Sequence[str]:
        """Get labels."""
        if self._labels is None:
            raise InvalidArgument("No labels were provided.")
        return self._labels



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module facilitates a Parallel version of AutoAttack.
"""
import logging
from copy import deepcopy
from typing import TYPE_CHECKING, Optional, Tuple, Union

import multiprocess
import numpy as np
from art.attacks import EvasionAttack
from art.attacks.evasion.auto_attack import AutoAttack
from art.config import ART_NUMPY_DTYPE
from art.utils import check_and_transform_label_format, get_labels_np_array

if TYPE_CHECKING:
    from art.utils import CLASSIFIER_TYPE

logger = logging.getLogger(__name__)


def run_attack(
    x: np.ndarray,
    y: np.ndarray,
    sample_is_robust: np.ndarray,
    attack: EvasionAttack,
    estimator: "CLASSIFIER_TYPE",
    norm: Union[int, float, str] = np.inf,
    eps: float = 0.3,
    gpu: bool = False,
    queue=None,
    **kwargs,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Run attack.

    :param x: An array of the original inputs.
    :param y: An array of the labels.
    :param sample_is_robust: Store the initial robustness of examples.
    :param attack: Evasion attack to run.
    :return: An array holding the adversarial examples.
    """
    # Attack only correctly classified samples
    if gpu:
        ident = multiprocess.current_process().ident
        gpu_id = queue.get()
        logger.info("%s: starting process on GPU %s", ident, gpu_id)
        estimator.set_device(f"cuda:{gpu_id}")
        attack.set_estimator(estimator)

    x_robust = x[sample_is_robust]
    y_robust = y[sample_is_robust]

    # Generate adversarial examples
    attack.verbose = False
    x_robust_adv = attack.generate(x=x_robust, y=y_robust, **kwargs)
    y_pred_robust_adv = estimator.predict(x_robust_adv)

    # Check and update successful examples
    rel_acc = 1e-4
    order = np.inf if norm == "inf" else norm
    assert isinstance(order, (int, float))
    norm_is_smaller_eps = (1 - rel_acc) * np.linalg.norm(
        (x_robust_adv - x_robust).reshape((x_robust_adv.shape[0], -1)), axis=1, ord=order
    ) <= eps

    if attack.targeted:
        samples_misclassified = np.argmax(y_pred_robust_adv, axis=1) == np.argmax(y_robust, axis=1)
    elif not attack.targeted:
        samples_misclassified = np.argmax(y_pred_robust_adv, axis=1) != np.argmax(y_robust, axis=1)
    else:  # pragma: no cover
        raise ValueError

    sample_is_not_robust = np.logical_and(samples_misclassified, norm_is_smaller_eps)

    x_robust[sample_is_not_robust] = x_robust_adv[sample_is_not_robust]
    x[sample_is_robust] = x_robust

    sample_is_robust[sample_is_robust] = np.invert(sample_is_not_robust)

    if gpu:
        logger.info("%s: Putting GPU %s back into queue.", ident, gpu_id)
        queue.put(gpu_id)

    return x, sample_is_robust


class JaticAutoAttack(AutoAttack):
    """
    JATIC defined extension of ART core AutoAttack
    """

    # Identify samples yet to have attack metadata identified
    SAMPLE_DEFAULT = -1
    # Identify samples misclassified therefore no attack metadata required
    SAMPLE_MISCLASSIFIED = -2

    def __init__(self, parallel: bool = False, gpu: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.parallel = parallel

        if gpu:
            from numba import cuda  # pylint: disable=C0415

            try:
                num_devices = len(cuda.gpus)
                logger.info("Found %s GPU devices.", num_devices)
                self.gpu = True
            except cuda.cudadrv.error.CudaSupportError as error:
                logger.debug("Cuda driver error: %s. Reverting to CPU.", error)
                self.gpu = False
        else:
            self.gpu = False

        if parallel:
            self.queue = multiprocess.Manager().Queue()
        self.best_attacks: np.ndarray = np.array([])

    def generate(self, x: np.ndarray, y: Optional[np.ndarray] = None, **kwargs) -> np.ndarray:
        """
        Generate adversarial samples and return them in an array.

        :param x: An array with the original inputs.
        :param y: Target values (class labels) one-hot-encoded of shape `(nb_samples, nb_classes)` or indices of shape
                  (nb_samples,). Only provide this parameter if you'd like to use true labels when crafting adversarial
                  samples. Otherwise, model predictions are used as labels to avoid the "label leaking" effect
                  (explained in this paper: https://arxiv.org/abs/1611.01236). Default is `None`.
        :param mask: An array with a mask broadcastable to input `x` defining where to apply adversarial perturbations.
                     Shape needs to be broadcastable to the shape of x and can also be of the same shape as `x`. Any
                     features for which the mask is zero will not be adversarially perturbed.
        :type mask: `np.ndarray`
        :return: An array holding the adversarial examples.
        """

        x_adv = x.astype(ART_NUMPY_DTYPE)
        if y is not None:
            y = check_and_transform_label_format(y, nb_classes=self.estimator_orig.nb_classes)

        if y is None:
            y = get_labels_np_array(self.estimator.predict(x, batch_size=self.batch_size))

        # Determine correctly predicted samples
        y_pred = self.estimator_orig.predict(x.astype(ART_NUMPY_DTYPE))
        sample_is_robust = np.argmax(y_pred, axis=1) == np.argmax(y, axis=1)

        # Set slots for images which have yet to be filled as SAMPLE_DEFAULT
        self.best_attacks = np.array([self.SAMPLE_DEFAULT] * len(x))
        # Set samples that are misclassified and do not need to be filled as SAMPLE_MISCLASSIFIED
        self.best_attacks[np.logical_not(sample_is_robust)] = self.SAMPLE_MISCLASSIFIED

        self.args = []
        # Untargeted attacks
        for attack in self.attacks:
            # Stop if all samples are misclassified
            if np.sum(sample_is_robust) == 0:
                break

            if attack.targeted:
                attack.set_params(targeted=False)

            if self.parallel:
                self.args.append(
                    (
                        deepcopy(x_adv),
                        deepcopy(y),
                        deepcopy(sample_is_robust),
                        deepcopy(attack),
                        deepcopy(self.estimator_orig),
                        deepcopy(self.norm),
                        deepcopy(self.eps),
                        deepcopy(self.gpu),
                        self.queue,
                    )
                )
            else:
                x_adv, sample_is_robust = run_attack(
                    x=x_adv,
                    y=y,
                    sample_is_robust=sample_is_robust,
                    attack=attack,
                    estimator=self.estimator,
                    norm=self.norm,
                    eps=self.eps,
                    gpu=self.gpu,
                    **kwargs,
                )
                # create a mask which identifies images which this attack was effective on
                # not including originally misclassified images
                atk_mask = np.logical_and(
                    np.array([i == self.SAMPLE_DEFAULT for i in self.best_attacks]), np.logical_not(sample_is_robust)
                )
                # update attack at image index with index of attack that was successful
                self.best_attacks[atk_mask] = self.attacks.index(attack)

        # Targeted attacks
        if self.targeted:
            # Labels for targeted attacks
            y_t = np.array([range(y.shape[1])] * y.shape[0])
            y_idx = np.argmax(y, axis=1)
            y_idx = np.expand_dims(y_idx, 1)
            y_t = y_t[y_t != y_idx]
            targeted_labels = np.reshape(y_t, (y.shape[0], -1))

            for attack in self.attacks:
                try:
                    attack.set_params(targeted=True)

                    for i in range(self.estimator.nb_classes - 1):
                        # Stop if all samples are misclassified
                        if np.sum(sample_is_robust) == 0:
                            break

                        target = check_and_transform_label_format(
                            targeted_labels[:, i], nb_classes=self.estimator.nb_classes
                        )
                        if self.parallel:
                            self.args.append(
                                (
                                    deepcopy(x_adv),
                                    deepcopy(target),
                                    deepcopy(sample_is_robust),
                                    deepcopy(attack),
                                    deepcopy(self.estimator_orig),
                                    deepcopy(self.norm),
                                    deepcopy(self.eps),
                                    deepcopy(self.gpu),
                                    self.queue,
                                )
                            )
                        else:
                            x_adv, sample_is_robust = run_attack(
                                x=x_adv,
                                y=target,
                                sample_is_robust=sample_is_robust,
                                attack=attack,
                                estimator=self.estimator,
                                norm=self.norm,
                                eps=self.eps,
                                gpu=self.gpu,
                                **kwargs,
                            )
                            # create a mask which identifies images which this attack was effective on
                            # not including originally misclassified images
                            atk_mask = np.logical_and(
                                np.array([i == self.SAMPLE_DEFAULT for i in self.best_attacks]),
                                np.logical_not(sample_is_robust),
                            )
                            # update attack at image index with index of attack that was successful
                            self.best_attacks[atk_mask] = self.attacks.index(attack)
                except ValueError as error:
                    logger.warning("Error completing attack: %s}", str(error))

        if self.parallel:
            if self.gpu:
                from numba import cuda  # pylint: disable=C0415

                devices = cuda.gpus
                for device in devices:
                    logger.info("Putting device %s into the queue.", device.id)
                    self.queue.put(device.id)
            with multiprocess.get_context("spawn").Pool() as pool:
                # Results come back in the order that they were issued
                results = pool.starmap(run_attack, self.args)
            perturbations = []
            is_robust = []
            for img_idx in range(len(x)):
                perturbations.append(np.array([np.linalg.norm(x[img_idx] - i[0][img_idx]) for i in results]))
                is_robust.append([i[1][img_idx] for i in results])
            best_attacks = np.argmin(np.where(np.invert(np.array(is_robust)), np.array(perturbations), np.inf), axis=1)
            x_adv = np.concatenate([results[best_attacks[img]][0][[img]] for img in range(len(x))])
            self.best_attacks = best_attacks
        return x_adv

    def __repr__(self) -> str:
        best_attack_meta = "\n".join(
            [
                f"image {i+1}: {str(self.args[idx][3])}" if idx != 0 else f"image {i+1}: n/a"
                for i, idx in enumerate(self.best_attacks)
            ]
        )
        auto_attack_meta = (
            f"AutoAttack(targeted={self.targeted}, parallel={self.parallel}, num_attacks={len(self.args)})"
        )
        return f"{auto_attack_meta}\nBestAttacks:\n{best_attack_meta}"



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements a JATIC compatible ART PyTorchYolo.
"""
from dataclasses import dataclass
from typing import Any, Sequence

import numpy as np
import torch
import yolov5
from art.estimators.object_detection.pytorch_yolo import PyTorchYolo
from maite.errors import InvalidArgument
from maite.protocols import HasDetectionPredictions, SupportsArray
from yolov5.utils.loss import ComputeLoss

from heart_library.utils import process_inputs_for_art


@dataclass
class JaticPyTorchObjectDetectionOutput:
    """
    Dataclass output of HEART PyTorch Classifier
    """

    scores: SupportsArray
    labels: SupportsArray
    boxes: SupportsArray


@dataclass
class HeartObjectDetectionMetadata:
    """
    Metadata dataclass of HEART YOLO Object Detector
    """

    model_name: str
    provider: str
    task: str = "Object Detection"


class Yolo(torch.nn.Module):
    """
    Wrapper for YOLO object detection models
    """

    def __init__(self, model, device):
        super().__init__()
        self.model = model
        self.model.hyp = {
            "box": 0.05,
            "obj": 1.0,
            "cls": 0.5,
            "anchor_t": 4.0,
            "cls_pw": 1.0,
            "obj_pw": 1.0,
            "fl_gamma": 0.0,
        }
        self.model.model.model.to(device)  # Required when using GPU

        self.compute_loss = ComputeLoss(self.model.model.model)

    def forward(self, x, targets=None):
        """
        Modified forward to facilitate computation of loss dict
        """
        if self.training:
            outputs = self.model.model.model(x)
            loss, loss_items = self.compute_loss(outputs, targets)
            loss_components_dict = {"loss_total": loss}
            loss_components_dict["loss_box"] = loss_items[0]
            loss_components_dict["loss_obj"] = loss_items[1]
            loss_components_dict["loss_cls"] = loss_items[2]
            return loss_components_dict
        return self.model(x)


class JaticPyTorchYolo(PyTorchYolo):  # pylint: disable=R0901
    """
    JATIC compatible extension of ART core PyTorchYolo
    """

    metadata: HeartObjectDetectionMetadata

    def __init__(self, labels: Sequence[str], model_path: str = "yolov5s.pt", **kwargs: Any):
        model = None
        try:
            model = yolov5.load(model_path)
        except Exception as load_exception:
            raise Exception("Yolov5 model was not successful loaded.") from load_exception

        model = Yolo(model, kwargs["device_type"])
        super().__init__(model=model, **kwargs)

        if labels is None:
            raise InvalidArgument("No labels were provided.")
        self._labels: Sequence[str] = labels
        self.metadata = HeartObjectDetectionMetadata(model_name="YOLOv5", provider="yolov5")

    def __call__(self, data: SupportsArray) -> HasDetectionPredictions:

        # convert to ART supported type
        images, _ = process_inputs_for_art(data, self._device)

        # make prediction
        output = self.predict(images)

        # convert back to JATIC supported type
        return JaticPyTorchObjectDetectionOutput(
            scores=np.concatenate([np.expand_dims(det["scores"], 0) for det in output]),
            boxes=np.concatenate([np.expand_dims(det["boxes"], 0) for det in output]),
            labels=np.concatenate([np.expand_dims(det["labels"], 0) for det in output]),
        )

    def get_labels(self) -> Sequence[str]:
        """Get labels."""
        if self._labels is None:
            raise InvalidArgument("No labels were provided.")
        return self._labels



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements a JATIC compatible ART Attack.
"""
from dataclasses import dataclass
from typing import Union

import numpy as np
from art.attacks import Attack, EvasionAttack, PoisoningAttack
from maite.protocols import SupportsArray

from heart_library.utils import process_inputs_for_art


@dataclass
class JaticEvasionAttackOutput:
    """
    Dataclass output JaticEvasionAttackOutput
    """

    adversarial_examples: SupportsArray
    adversarial_patch: SupportsArray


@dataclass
class JaticPoisonAttackOutput:
    """
    Dataclass output JaticEvasionAttackOutput
    """

    poisoning_examples: SupportsArray
    poisoning_labels: SupportsArray


class JaticAttack:  # pylint: disable=R0901
    """
    Wrapper for JATIC compatible attacks
    """

    def __init__(self, attack: Attack):
        self.attack = attack

    def run_attack(self, data: SupportsArray, **kwargs) -> Union[JaticEvasionAttackOutput, JaticPoisonAttackOutput]:
        """
        Run the attack.. depending on the attack, call different func.. depending on the attack, different output
        .. metrics?
        """

        # convert to ART supported data types
        x, y = process_inputs_for_art(data, self.attack.estimator.device)

        if y is not None and not isinstance(y, np.ndarray):
            # check if y is a list of dicts, as per object detection
            if not all(isinstance(i, dict) for i in y):
                y = np.array(y)

        # run the attack
        if isinstance(self.attack, EvasionAttack):
            results = self.attack.generate(x, y, **kwargs)

            if hasattr(self.attack, "apply_patch"):
                patch = results
                adv_images = self.attack.apply_patch(x=x, scale=0.4)
                return JaticEvasionAttackOutput(adversarial_examples=adv_images, adversarial_patch=patch)
            # convert to JATIC supported data types
            return JaticEvasionAttackOutput(adversarial_examples=results, adversarial_patch=np.array([]))

        if isinstance(self.attack, PoisoningAttack):
            results = self.attack.poison(x, y, **kwargs)
            # convert to JATIC supported data types
            return JaticPoisonAttackOutput(**results)

        raise Exception("Invalid attack.")



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Module implementing varying metrics for assessing model robustness. These fall mainly under two categories:
attack-dependent and attack-independent.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import logging
from typing import Any, Dict, Sequence, Tuple

import numpy as np
import numpy.linalg as la
from maite.protocols import (ArrayLike, HasLogits, HasProbs, HasScores,
                             ImageClassifier, SupportsArray)
from numpy.typing import NDArray

from heart_library.utils import process_inputs_for_art

logger = logging.getLogger(__name__)


class RobustnessBiasMetric:
    """
    A metric which describes Robustness Bias for features
    of datasets. Currently supports only classification tasks.
    """
    def __init__(self):
        self._state = {}

    def reset(self):
        """
        Reset the metric to default values.
        """
        self._state = {}

    def update(
        self,
        classifier: ImageClassifier,
        device: str,
        data: SupportsArray,
        attack_out: NDArray,
        labels: NDArray = np.array(None),
        norm_type: int = 2,
        interval: int = 100,
    ):
        """
        Updates the metric value. Takes as input:
        :param classifier: The image classifier
        :param device: The device on which to compute the metric
        :param data: The clean sample data
        :param attack_out: The adversarial sample data
        :param labels: The classification labels
        :param norm_type: The norm to use when calculating distance
        """

        y: Any = np.array(None)
        if labels.all() is None:
            # labels not explicitly provided, assume they are in data
            x, y = process_inputs_for_art(data, device)
        else:
            x, _ = process_inputs_for_art(data, device)
            if isinstance(y, int):
                y = [labels]
            else:
                y = labels

        if isinstance(x, Sequence):
            assert x[0].shape == attack_out.shape
        else:
            assert x.shape == attack_out.shape

        if isinstance(y, np.ndarray) and len(y.shape) > 1:
            y = np.argmax(y, axis=1)

        errors = np.linalg.norm((x - attack_out).reshape(x.shape[0], -1), ord=norm_type, axis=1)

        orig_result = classifier(x)
        attack_result = classifier(attack_out)

        success: NDArray
        if isinstance(orig_result, HasLogits) and isinstance(attack_result, HasLogits):
            success = (np.argmax(orig_result.logits, axis=1) != np.argmax(attack_result.logits, axis=1)).astype(int)
        elif isinstance(orig_result, HasProbs) and isinstance(attack_result, HasProbs):
            success = (np.argmax(orig_result.probs, axis=1) != np.argmax(attack_result.probs, axis=1)).astype(int)
        elif isinstance(orig_result, HasScores) and isinstance(attack_result, HasScores):
            success = (np.argmax(orig_result.scores, axis=1) != np.argmax(attack_result.scores, axis=1)).astype(int)
        else:
            raise ValueError

        taus = np.linspace(0, max(errors) + 1, interval)
        series: Dict = {}
        for tau in taus:
            for label in range(len(classifier.get_labels())):
                idxs_of_label = np.argwhere(np.array(y) == label).ravel()
                idxs_of_label_success = np.argwhere(success[idxs_of_label] == 1).ravel()
                idxs_of_label = idxs_of_label[idxs_of_label_success]
                errors_of_label = errors[idxs_of_label]
                idx_error_greater_tau = np.argwhere(errors_of_label > tau).ravel()
                if len(errors_of_label) != 0:
                    proportion = len(idx_error_greater_tau) / len(errors_of_label)
                    if label in series:
                        series[label].append([tau, proportion])
                    else:
                        series[label] = [[tau, proportion]]

        self._state = series

    def compute(self) -> Tuple[ArrayLike, ArrayLike, ArrayLike]:
        """
        Returns the computed metric
        in Tuple (clean_accuracy, robust_accuracy, average_perturbation)
        """
        return self._state

    def to(self, device: Any):  # pylint: disable=C0103,W0613
        """Unused protocol"""
        return self


class AccuracyPerturbationMetric:
    """
    A metric for easily calculating the clean and robust accuracy
    as well as the perturbation between clean and adversarial input.
    """

    def __init__(self):
        self._state = (np.array([0.0]), np.array([0.0]), np.array([0.0]))

    def reset(self):
        """
        Reset the metric to default values.
        """
        self._state = (np.array([0.0]), np.array([0.0]), np.array([0.0]))

    def update(
        self,
        classifier: ImageClassifier,
        device: str,
        data: SupportsArray,
        attack_out: NDArray,
        labels: NDArray = np.array(None),
        norm_type: int = 2,
        acc_type: str = "robust",
    ):
        """
        Updates the metric value. Takes as input:
        :param classifier: The image classifier
        :param device: The device on which to compute the metric
        :param data: The clean sample data
        :param attack_out: The adversarial sample data
        :param labels: The classification labels
        :param norm_type: The norm to use when calculating distance
        :param acc_type: The type of accuracy to calculate. Choice of "adversarial" or "robust".
            - Robust accuracy is the accuracy of the model on all samples
            - Adversarial accuracy is the accuracy of the model only samples which were
              correctly predicted in the non-adversarial scenario
        """
        y: Any = np.array(None)
        if labels.all() is None:
            # labels not explicitly provided, assume they are in data
            x, y = process_inputs_for_art(data, device)
        else:
            x, _ = process_inputs_for_art(data, device)
            if isinstance(y, int):
                y = [labels]
            else:
                y = labels

        if isinstance(x, Sequence):
            assert x[0].shape == attack_out.shape
        else:
            assert x.shape == attack_out.shape

        orig_result = classifier(x)
        attack_result = classifier(attack_out)

        if isinstance(orig_result, HasLogits) and isinstance(attack_result, HasLogits):
            y_orig = np.argmax(orig_result.logits, axis=1)
            y_pred = np.argmax(attack_result.logits, axis=1)
        elif isinstance(orig_result, HasProbs) and isinstance(attack_result, HasProbs):
            y_orig = np.argmax(orig_result.probs, axis=1)
            y_pred = np.argmax(attack_result.probs, axis=1)
        elif isinstance(orig_result, HasScores) and isinstance(attack_result, HasScores):
            y_orig = np.argmax(orig_result.scores, axis=1)
            y_pred = np.argmax(attack_result.scores, axis=1)
        else:
            raise ValueError

        if isinstance(y, np.ndarray) and len(y.shape) > 1:
            y = np.argmax(y, axis=1)

        y_corr = y_orig == y

        clean_acc = np.sum(y_corr) / len(y_orig)
        if acc_type == "adversarial":
            attack_acc = np.sum((y_pred == y_orig) & y_corr) / np.sum(y_corr)
        elif acc_type == "robust":
            attack_acc = np.mean(y_pred == y)

        idxs = y_pred != y
        avg_perts = 0.0
        perts_norm = la.norm((attack_out - x).reshape(x.shape[0], -1), ord=norm_type, axis=1)
        perts_norm = perts_norm[idxs]
        avg_perts = np.mean(perts_norm)

        self._state = (clean_acc, attack_acc, avg_perts)

    def compute(self) -> Tuple[ArrayLike, ArrayLike, ArrayLike]:
        """
        Returns the computed metric
        in Tuple (clean_accuracy, robust_accuracy, average_perturbation)
        """
        return self._state

    def to(self, device: Any):  # pylint: disable=C0103,W0613
        """Unused protocol"""
        return self



"""
Module providing metrics
"""
from heart_library.metrics.metrics import (AccuracyPerturbationMetric,
                                   RobustnessBiasMetric)

__all__ = ("AccuracyPerturbationMetric", "RobustnessBiasMetric")



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Utility methods for converting data types to ART compatible versions.
"""
from collections import UserDict
from typing import Any, Optional, Sequence, Tuple, Union

import numpy as np
import torch as tr
from maite._internals.interop.utils import is_pil_image
from maite.protocols import ArrayLike, HasDataImage, HasDataLabel, is_list_of_type, is_typed_dict
from numpy.typing import NDArray


def process_inputs_for_art(
    data: Union[HasDataImage, Sequence[ArrayLike], ArrayLike], device
) -> Tuple[NDArray, Optional[Any]]:
    """Convert JATIC supported data to ART supported data"""

    images: Any = None
    labels: Any = None

    # If single PIL image, list of one and handle below
    if is_pil_image(data):
        data = [data]

    # If single array, try to convert to tensor and stack and just return
    if not isinstance(data, (dict, UserDict, Sequence)):
        images = tr.as_tensor(data).to(device)
        if images.ndim == 3:
            images = images.unsqueeze(0)
        return images.cpu().numpy(), None

    # Extract any data out of a dictionary
    labels = None
    if isinstance(data, (dict, UserDict)):
        assert is_typed_dict(data, HasDataImage), "Dictionary data must contain 'image' key."
        images = data["image"]

        if not isinstance(images, Sequence):
            images = [images]

        if is_typed_dict(data, HasDataLabel):
            labels = data["label"]

            if not isinstance(labels, Sequence):
                labels = [labels]

            if isinstance(labels, Sequence) and all(isinstance(label, tr.Tensor) for label in labels):
                labels = [label.cpu().numpy()[0] for label in labels]
    else:
        images = data

    assert isinstance(images, Sequence)

    # If list of PIL images, convert to tensor
    if is_pil_image(images[0]):
        images = [tr.as_tensor(np.array(image, np.float32) / 255).permute((2, 0, 1)).contiguous() for image in images]

        if isinstance(images, Sequence):
            assert isinstance(images[0], tr.Tensor), f"Invalid type {type(images[0])}"
            images = [tr.as_tensor(i).to(device) for i in images]
        else:
            assert isinstance(images, tr.Tensor)
            images = images.to(device)

    # If list of numpy arrays, convert to tensor
    if is_list_of_type(images, np.ndarray):  # type: ignore
        if len(images[0].shape) < 4:
            images = [tr.as_tensor(i).to(device) for i in images]
        else:
            images = tr.as_tensor(images[0]).to(device)

    # If list of numpy arrays
    if is_list_of_type(labels, np.ndarray):  # type: ignore
        if len(labels[0].shape) >= 2:
            labels = labels[0]

    # For list of Tensors, try to stack into a single tensor
    if is_list_of_type(images, tr.Tensor):
        first_item = images[0]
        shape_first = first_item.shape  # type: ignore

        def gen():
            return (shape_first == item.shape for item in images)

        if all(gen()):
            images = tr.stack(images)

    assert is_list_of_type(images, tr.Tensor) or isinstance(images, tr.Tensor)
    if is_list_of_type(images, tr.Tensor):
        images = tr.cat(images).cpu().numpy()
    if isinstance(images, tr.Tensor):
        images = images.cpu().numpy()
    return images, labels



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
ART attack Protocols
"""
from typing import Protocol, Union, runtime_checkable

from maite.protocols import SupportsArray


@runtime_checkable
class HasEvasionAttackResult(Protocol):
    """
    Scores are predictions for either an image or detection box.

    Attributes
    ----------
    adversarial_examples : SupportsArray
        Adversarial sample computed for an evasion attack.

    Examples
    --------

    """

    adversarial_examples: SupportsArray


@runtime_checkable
class HasPoisonAttackResult(Protocol):
    """
    Scores are predictions for either an image or detection box.

    Attributes
    ----------
    poisoning_examples : SupportsArray
        Poisoned samples computed for a poisoning attack.

    poisoning_labels : SupportsArray
        Poisoned labels computed for a poisoning attack.

    Examples
    --------

    """

    poisoning_examples: SupportsArray
    poisoning_labels: SupportsArray


@runtime_checkable
class Attack(Protocol):
    """
    An attack that takes in images and optionally labels and returns attack result e.g. adversarial image

    Methods
    -------
    run_attack(data: SupportsArray, ) -> Union[HasLogits, HasProbs, HasScores]
        Run inference on the data.

    Examples
    --------

    """

    def run_attack(self, data: SupportsArray, **kwargs) -> Union[HasEvasionAttackResult, HasPoisonAttackResult]:
        """Returns the labels for the model."""
        ...
